/**
 * @file botservos.h
 * This defines the BotServos class
 */
#ifndef __BOTSERVOS_H
#define __BOTSERVOS_H

#if defined(__cplusplus)
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/thread.h>
#include <propeller.h>

#ifndef dprintf
#define dprintf
#endif

/* These numbers define servo stop pulse width */
static const int LEFT_CENTER = 1500;
static const int RIGHT_CENTER = 1500;

#ifndef STACK_SIZE 
#define STACK_SIZE 32
#endif

/* stacks for COGs */
static int LeftServoStack  [STACK_SIZE];
static int RightServoStack [STACK_SIZE];

/* per-thread library variables */
static _thread_state_t LeftThreadData;
static _thread_state_t RightThreadData;

/* WheelServo COG variables */
static int LeftServoSpeed;
static int RightServoSpeed;
static int LeftRotation;
static int RightRotation;

static int MaxServoSpeed = 100;

static int mLeftServoPin;
static int mRightServoPin;

/*
 * Servo COG function
 */
void LeftServoThread(void *arg __attribute__((unused)))
{
    int pinmask = (1<<mLeftServoPin);
    int center = LEFT_CENTER;
    int ustime = CLKFREQ/1000000;

    DIRA |= pinmask;

    for(;;) {
        /* start with signal off */
        OUTA &= ~pinmask;

        if(LeftServoSpeed == 0) {
            waitcnt(ustime*center+CNT);
            LeftRotation = 0;
            continue;
        }

        if(LeftServoSpeed > MaxServoSpeed)
            LeftServoSpeed = MaxServoSpeed;
        if(LeftServoSpeed < -MaxServoSpeed)
            LeftServoSpeed = -MaxServoSpeed;

        /* move servo */
        LeftRotation = (-LeftServoSpeed/2*center)/MaxServoSpeed+center;
        waitcnt(ustime*center+CNT);
        OUTA |=  pinmask;
        waitcnt(ustime*LeftRotation+CNT);
        OUTA ^=  pinmask;
    }
}

/*
 * Servo COG function
 */
static void RightServoThread(void *arg __attribute__((unused)))
{
    int pinmask = (1<<mRightServoPin);
    int center = RIGHT_CENTER;
    int ustime = CLKFREQ/1000000;

    DIRA |= pinmask;

    for(;;) {
        /* start with signal off */
        OUTA &= ~pinmask;

        if(RightServoSpeed == 0) {
            waitcnt(ustime*center+CNT);
            RightRotation = 0;
            continue;
        }

        if(RightServoSpeed > MaxServoSpeed)
            RightServoSpeed = MaxServoSpeed;
        if(RightServoSpeed < -MaxServoSpeed)
            RightServoSpeed = -MaxServoSpeed;

        /* move servo */
        RightRotation = (RightServoSpeed/2*center)/MaxServoSpeed+center;
        waitcnt(ustime*center+CNT);
        OUTA |=  pinmask;
        waitcnt(ustime*RightRotation+CNT);
        OUTA ^=  pinmask;
    }
}


static void startServoThreads(void)
{
    int n;
    n = _start_cog_thread(LeftServoStack + STACK_SIZE, LeftServoThread, NULL, &LeftThreadData);
    dprintf("Left Servo COG thread %d\n", n);
    n = _start_cog_thread(RightServoStack + STACK_SIZE, RightServoThread, NULL, &RightThreadData);
    dprintf("Right Servo COG thread %d\n", n);
}

#if defined(__cplusplus)
};
#endif

class BotServos
{
public:
    BotServos(int leftPin, int rightPin)
    {
        ::mLeftServoPin = leftPin;
        ::mRightServoPin = rightPin;
    }

    void botStop(void)
    {
        ::LeftServoSpeed = 0;
        ::RightServoSpeed = 0;
    }
    
    void botForward(int rate)
    {
        ::LeftServoSpeed = rate;
        ::RightServoSpeed = rate;
    }
    
    void botReverse(int rate)
    {
        ::LeftServoSpeed = -rate;
        ::RightServoSpeed = -rate;
    }
    
    void botLeftTurn(int rate)
    {
        ::LeftServoSpeed = 0;
        ::RightServoSpeed = rate;
    }
    
    void botRightTurn(int rate)
    {
        ::LeftServoSpeed = rate;
        ::RightServoSpeed = 0;
    }
    
    void botLeftSpin(int rate)
    {
        ::LeftServoSpeed = rate;
        ::RightServoSpeed = -rate;
    }
    
    void botRightSpin(int rate)
    {
        ::LeftServoSpeed = -rate;
        ::RightServoSpeed = rate;
    }
    
    /**
     * Main program function.
     */
    int test(void)
    {
        int move = 's';
        int speed = 100;
    
        /* keep QuickStart TX quiet. */
        DIRA |= (1<<30);
    
        setbuf(stdout, 0);
        setbuf(stdin, 0);
        waitcnt(500*(CLKFREQ/1000)+CNT);
    
        ::startServoThreads();
        botStop();
    
        for(;;) {
            putchar(0);
            dprintf("PropBOE Servo Test\n");
            dprintf("Last Command '%c'. Speed %d\n", move, speed);
            dprintf("LeftRotation = %d ",LeftRotation);
            dprintf("RightRotation = %d\n",RightRotation);
            dprintf("LeftServoSpeed = %d ",LeftServoSpeed);
            dprintf("RightServoSpeed = %d\n",RightServoSpeed);
            dprintf("\n$ ");
            move = getchar();
            getchar();
            switch(move) {
                case 'f':
                    botForward(speed);
                    dprintf("forward %d\n", speed);
                break;
                case 'b':
                    botReverse(speed);
                    dprintf("backward %d\n", speed);
                break;
                case 'l':
                    botLeftTurn(speed);
                    dprintf("left  turn %d\n", speed);
                break;
                case 'r':
                    botRightTurn(speed);
                    dprintf("right turn %d\n", speed);
                break;
                case 'L':
                    botLeftSpin(speed);
                    dprintf("left  spin %d\n", speed);
                break;
                case 'R':
                    botRightSpin(speed);
                    dprintf("right spin %d\n", speed);
                break;
                case 's':
                    botStop();
                    dprintf("stop\n");
                break;
                case '+':
                    speed += 100;
                    dprintf("speed %d\n", speed);
                break;
                case '-':
                    speed -= 100;
                    dprintf("speed %d\n", speed);
                break;
            }
    
        }
    
        return 0;
    }

};

#endif