/**
 * @file servo.h
 * This defines a very simple continuous rotation Servo class.
 * It is not designed for servo positioning.
 */
#ifndef __SERVO_H
#define __SERVO_H

#if defined(__cplusplus)
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/thread.h>
#include <propeller.h>

#ifndef dprintf
#define dprintf
#endif

/* These numbers define servo stop pulse width */
static const int SERVO_CENTER = 1500;

#ifndef STACK_SIZE 
#define STACK_SIZE 32
#endif

/* stacks for COGs */
static int ServoStack  [STACK_SIZE];

/* per-thread library variables */
static _thread_state_t ThreadData;

/* WheelServo COG variables */
static int ServoSpeed;
static int Rotation;

static int MaxServoSpeed = 100;

static int mServoPin;

/*
 * Servo COG function
 */
void ServoThread(void *arg __attribute__((unused)))
{
    int pinmask = (1<<mServoPin);
    int center = SERVO_CENTER;
    int ustime = CLKFREQ/1000000;

    DIRA |= pinmask;

    for(;;) {
        /* start with signal off */
        //OUTA &= ~pinmask;

        if(ServoSpeed == 0) {
            waitcnt(ustime*center+CNT);
            Rotation = 0;
            continue;
        }

        if(ServoSpeed > MaxServoSpeed)
            ServoSpeed = MaxServoSpeed;
        if(ServoSpeed < -MaxServoSpeed)
            ServoSpeed = -MaxServoSpeed;

        /* move servo */
        Rotation = (-ServoSpeed/2*center)/MaxServoSpeed+center;
        waitcnt(ustime*center+CNT);
        OUTA |=  pinmask;
        waitcnt(ustime*Rotation+CNT);
        OUTA ^=  pinmask;
    }
}

static void startServoThread(void)
{
    int n;
    n = _start_cog_thread(ServoStack + STACK_SIZE, &ServoThread, NULL, &ThreadData);
    dprintf("Servo COG thread %d\n", n);
}

#if defined(__cplusplus)
};
#endif

class Servo
{
public:
    Servo(int pin)
    {
        ::mServoPin = pin;
    }

    void startThread(void)
    {
        ::startServoThread();
    }

    void stopServo(void)
    {
        ::ServoSpeed = 0;
    }
    
    void forward(int rate)
    {
        ::ServoSpeed = rate;
    }
    
    void reverse(int rate)
    {
        ::ServoSpeed = -rate;
    }
    
    /**
     * Main program function.
     */
    int test(void)
    {
        int move = 's';
        int speed = 100;
    
        /* keep QuickStart TX quiet. */
        DIRA |= (1<<30);
    
        setbuf(stdout, 0);
        setbuf(stdin, 0);
        waitcnt(500*(CLKFREQ/1000)+CNT);
    
        startThread();
        stopServo();
    
        for(;;) {
            putchar(0);
            dprintf("PropBOE Servo Test\n");
            dprintf("Last Command '%c'. Speed %d\n", move, speed);
            dprintf("Rotation = %d ",Rotation);
            dprintf("ServoSpeed = %d ",ServoSpeed);
            dprintf("\n$ ");
            move = getchar();
            getchar();
            switch(move) {
                case 'f':
                    forward(speed);
                    dprintf("forward %d\n", speed);
                break;
                case 'b':
                    reverse(speed);
                    dprintf("backward %d\n", speed);
                break;
                case 's':
                    stopServo();
                    dprintf("stop\n");
                break;
                case '+':
                    speed += 100;
                    dprintf("speed %d\n", speed);
                break;
                case '-':
                    speed -= 100;
                    dprintf("speed %d\n", speed);
                break;
            }
    
        }
    
        return 0;
    }
#endif
};

#endif